function geteigm(A::SymTridiagonal; useArnoldi::Bool=false)
  if useArnoldi
    converged, eig = geteigmArnoldi(A);
    return converged, eig
  else
    converged, eig = geteigmLA(A);
    return converged, eig
  end
end

function geteigmLA(A::SymTridiagonal)
  eig = eigmax(A);
  return false, eig
end

function geteigmArnoldi(A::SymTridiagonal)
  #Attempt to use Arnoldi method failed
  restarts = 100
  d, h = ArnoldiMethod.partialschur(A, nev=1, which=LR(), restarts = restarts, tol=1e-6)
  maxtries = 3
  tries = 1
  while !h.converged && tries < maxtries
    restarts *= 10
    d, h = ArnoldiMethod.partialschur(A, nev=1, which=LR(), restarts=restarts, tol=1e-6)
    tries += 1
    #println("restarts: ",restarts)
  end
  #println("converged: ",h.converged)
  eigm = real(d.eigenvalues[1])

  if !h.converged
    return true, 0.0
  else
    return false, eigm
  end

end
