using CairoMakie

function coef_grid_test(;X::AbstractFloat=1.0)
  s = 0.5;
  Nζ = 10;
  Nα = 10;

  ζs = (0:Nζ-1) .* (2π/Nζ);
  αvec = (0:Nα-1) .* (2π/Nα);

  vmec = VMEC.readVmecWout("../example/wout_ku4.nc");
  vmecSurface = VmecSurface(s, vmec);

  c2_coefs, κ_coefs, bsuppar = ballooning_coefficient_grid(vmecSurface, ζs, αvec)
  c2 = c2_coefs[:, :, 1] + c2_coefs[:, :, 2] .* X + c2_coefs[:, :, 3] .* X^2;
  κ  = κ_coefs[:, :, 1]  + κ_coefs[:, :, 2]  .* X;

  c2err      = zeros(Nα, Nζ);
  κerr       = zeros(Nα, Nζ);
  bsupparerr = zeros(Nα, Nζ);

  c2summodosd = zeros(Nα, Nζ);
  κsummodosd = zeros(Nα, Nζ);
  bsupparsummodosd = zeros(Nα, Nζ);


  vmecCoords = VmecCoordinates(vmecSurface.s, 0, 0)
  pestCoords = PestFromVmec()(vmecCoords, vmecSurface)
  ψpest = pestCoords.ψ
  for iα = 1:Nα, iζ = 1:Nζ
    ζ = ζs[iζ];
    pestCoords  = PestCoordinates(ψpest, αvec[iα], -(ζ - X));
    vmecCoords  = VmecFromPest()(pestCoords,vmecSurface)

    θ = vmecCoords.θ;

    c2i = [0.0];
    κi = [0.0];
    bsuppari = [0.0];
    b2i = [0.0];
    summodosd(vmecSurface, ζ-X, θ, 1, [X], c2i, κi, bsuppari, b2i)

    c2summodosd[iα, iζ] = c2i[1]/b2i[1];
    κsummodosd[iα, iζ] = κi[1];
    bsupparsummodosd[iα, iζ] = bsuppari[1];
  end

  println("c2 relative error = $(norm(c2-c2summodosd)/norm(c2))");
  println("κ relative error = $(norm(κ - κsummodosd)/norm(κ))");
  println("bsuppar relative error = $(norm(bsuppar - bsupparsummodosd)/norm(bsuppar))");

end

function coeffs_time_test(N, vmecSurface, s, init_zeta, init_theta, npt, coeff_data, params)
  for ii = 1:N
    coeffs(vmecSurface, s, init_zeta, init_theta, npt, coeff_data, params)
  end
end

function coeffs_interp_time_test(N, bci2, init_zeta, init_alpha, coeff_data_interp)
  for ii = 1:N
    coeffs(bci2, init_zeta, init_alpha, coeff_data_interp)
  end
end


function BallooningCoefInterp_test()
  # First, test the transformation matrices
  s = 0.6;
  Nζ = 80;
  Nθ = 40;

  AF     = FourierGridMat(Nθ; backward=false);
  AFinvT = FourierGridMat(Nθ; backward=true);

  AC     = ChebyshevGridMat(Nζ; backward=false);
  ACinvT = ChebyshevGridMat(Nζ; backward=true);

  println("Fourier test   norm(AFinvT'*AF-I) = $(norm(AFinvT'*AF-I))");
  println("Chebyshev test norm(ACinvT'*AC-I) = $(norm(ACinvT'*AC-I))");


  # Load VMEC
  vmec = VMEC.readVmecWout("../example/wout_ku4.nc");
  vmecSurface = VmecSurface(s, vmec);
  nfp = vmecSurface.nfp;
  println("nfp = $(nfp), mpol = $(vmecSurface.mpol), ntor = $(vmecSurface.mpol)")


  # Interpolate
  bci = BallooningCoefInterp(vmecSurface, Nζ, Nθ)

  ζs = (π/nfp) .* (-cos.(LinRange(0, π, Nζ+1)) .+ 1);
  θinitvec = (0:2Nθ) .* (2π/(2Nθ+1));
  c2_interp, κ_interp, bsuppar_interp = eval_coefs(bci, θinitvec, ζs);
  c2, κ, bsuppar = ballooning_coefficient_grid(vmecSurface, ζs, θinitvec);

  #Check error on grid
  println("\nerror in c2 interpolation on grid:      $(norm(c2_interp - c2))")
  println("error in κ interpolation on grid:       $(norm(κ_interp - κ))")
  println("error in bsuppar interpolation on grid: $(norm(bsuppar_interp - bsuppar))")


  # Check error on double grid
  ζs = (π/nfp) .* (-cos.(LinRange(0, π, 2Nζ+1)) .+ 1);
  θinitvec = (0:4Nθ+1) .* (2π/(4Nθ+2));
  c2_interp, κ_interp, bsuppar_interp = eval_coefs(bci, θinitvec, ζs);
  c2, κ, bsuppar = ballooning_coefficient_grid(vmecSurface, ζs, θinitvec);

  println("\nrelative error in c2 interpolation on double grid:      $(norm(c2_interp - c2) / norm(c2))")
  println("relative error in κ interpolation on double grid:       $(norm(κ_interp - κ) / norm(κ))")
  println("relative error in bsuppar interpolation on double grid: $(norm(bsuppar_interp - bsuppar) / norm(bsuppar))")
  println("norm(c2_interp - c2) = $(norm(c2_interp - c2)), norm(c2) = $(norm(c2))");
  println("norm(κ_interp - κ) = $(norm(κ_interp - κ)), norm(κ) = $(norm(κ))");
  println("norm(bsuppar_interp - bsuppar) = $(norm(bsuppar_interp - bsuppar)), norm(bsuppar) = $(norm(bsuppar))");

  dfα = norm(map_α_inv(bci, map_α(bci, θinitvec)) - θinitvec);
  println("Norm of error of fα⁻¹(fα(α)) - α: $dfα");

  # Check that we map between surfaces correctly
  iota = vmecSurface.iota[1];
  c2_bot, κ_bot, bsuppar_bot = eval_coefs(bci, map_α(bci, θinitvec), [0.0]);
  c2_top, κ_top, bsuppar_top = eval_coefs(bci, θinitvec, [2π/nfp]);

  println("relative norm of c2_bot - c2_top is           $(norm(c2_bot-c2_top)/norm(c2_bot))")
  println("relative norm of κ_bot - κ_top is             $(norm(κ_bot-κ_top)/norm(κ_bot))")
  println("relative norm of bsuppar_bot - bsuppar_top is $(norm(bsuppar_bot-bsuppar_top)/norm(bsuppar_bot))")


  # Check that we get the right p, q, r
  init_zeta = 0.23;
  init_theta = 0.263;
  npt = 1000;
  L = 5π;

  vmecCoords = VmecCoordinates(vmecSurface.s, init_theta, init_zeta)
  pestCoords = PestFromVmec()(vmecCoords, vmecSurface)
  init_alpha = pestCoords.α

  coeff_data = coeffs_struct(npt);
  coeff_data.x .= LinRange(-L, L, npt);

  coeff_data_interp = deepcopy(coeff_data)

  nwells = 1;
  geom_input = true;
  tokamak_input=false;
  mode=1;
  vmec_data = VmecData(vmec)
  (r0, amin, beta0, b0_v) = order_input(vmec_data);
  params = BalloonParams(nwells, geom_input, tokamak_input, mode, r0, amin, beta0, b0_v)

  coeffs(vmecSurface, s, init_zeta, init_theta, npt, coeff_data, params)

  bci2 = BallooningCoefInterp(vmecSurface, Nζ, Nθ, params)
  coeffs(bci2, init_zeta, init_alpha, coeff_data_interp)

  x = coeff_data.x;
  q = coeff_data.q;
  p = coeff_data.p;
  r = coeff_data.r;
  q_interp = coeff_data_interp.q;
  p_interp = coeff_data_interp.p;
  r_interp = coeff_data_interp.r;

  println("q relative error = $(norm(q-q_interp)/norm(q))")
  println("p relative error = $(norm(p-p_interp)/norm(p))")
  println("r relative error = $(norm(r-r_interp)/norm(r))")


  p12 = plot(x, p, color=:blue, label="coeffs");
  plot!(x,p_interp, color=:red, label="interp");
  display(p12);

  p13 = plot(x, r, color=:blue, label="coeffs");
  plot!(x,r_interp, color=:red, label="interp");
  display(p13);

  N = 10
  t = @time coeffs_time_test(N, vmecSurface, s, init_zeta, init_theta, npt, coeff_data, params);
  println("coeffs time = $(t)")
  t = @time coeffs_interp_time_test(N, bci2, init_zeta, init_alpha, coeff_data_interp);
  println("interpolated coeffs time = $(t)")

  reltols = [1e-1, 5e-5, 1e-8];
  for reltol in reltols
    bci3 = BallooningCoefInterp(vmecSurface; reltol)
    bci4 = BallooningCoefInterp(vmecSurface, get_Nζ(bci3), get_Nα(bci3));
    println("Adaptive bci: reltol = $reltol, Nα = $(get_Nα(bci3)), Nζ = $(get_Nζ(bci3))")
    println("norm(bci3.c2-bci4.c2) + norm(bci3.κ-bci4.κ) + norm(bci3.bsuppar -bci4.bsuppar) = $(norm(bci3.c2-bci4.c2) + norm(bci3.κ-bci4.κ) + norm(bci3.bsuppar -bci4.bsuppar))")
  end
end

function test_coeffs_derivs!()
  s = 0.5;
  r = [0.1, 0.2];
  ϵ = 1e-7;
  dr = ϵ*randn(2);
  r2 = r + dr;

  # Check derivative matrices:
  Nζ = 4;
  Nα = 3;
  x = 0.25 .* (-4:4);
  dx = ones(9); dx[end] = -1;
  dx = dx .* ϵ;
  x2 = x + dx;
  θ = acos.(x);
  θ2 = acos.(x2);
  AC = ChebyshevMat(θ, Nζ)
  dAC = dChebyshevMat(θ, Nζ)
  ddAC = ddChebyshevMat(θ, Nζ)
  AC2 = ChebyshevMat(θ2, Nζ)
  dAC2 = dChebyshevMat(θ2, Nζ)

  println("CHEBYSHEV FINITE DIFFERENCE CHECK:
  AC: $(norm(AC2 - AC - Diagonal(dx)*dAC) / ϵ)
  dAC: $(norm(dAC2 - dAC - Diagonal(dx)*ddAC) / ϵ)
")

  α = (0:8) .* (2π/9)
  dα = ϵ .* ones(length(α));
  α2 = α + dα
  AF = FourierMat(α, Nζ)
  dAF = dFourierMat(α, Nζ)
  ddAF = ddFourierMat(α, Nζ)
  AF2 = FourierMat(α2, Nζ)
  dAF2 = dFourierMat(α2, Nζ)
  println("FOURIER FINITE DIFFERENCE CHECK:
  AF: $(norm(AF2 - AF - Diagonal(dα)*dAF) / ϵ)
  dAF: $(norm(dAF2 - dAF - Diagonal(dα)*ddAF) / ϵ)
  ")



  # Load VMEC
  vmec = VMEC.readVmecWout("../example/wout_ku4.nc");
  vmecSurface = VmecSurface(s, vmec);
  # bci = BallooningCoefInterp(vmecSurface, 5, 5)
  bci = BallooningCoefInterp(vmecSurface)
  nfp = vmecSurface.nfp;
  println("nfp = $(nfp), mpol = $(vmecSurface.mpol), ntor = $(vmecSurface.mpol)")

  # Check ddeval_coefs
  ζ = 0.2:0.1:0.3
  α = 0:1:3
  dζ = ϵ * randn(length(ζ));
  dα = ϵ * zeros(length(α));
  ζ2 = ζ + dζ;
  α2 = α + dα

  c2, κ, bsuppar, dc2, dκ, dbsuppar, ddc2, ddκ, ddbsuppar = ddeval_coefs(bci, α, ζ)
  c22, κ2, bsuppar2, dc22, dκ2, dbsuppar2, ddc22, ddκ2, ddbsuppar2 = ddeval_coefs(bci, α2, ζ2)

  c2_err = zeros(length(α), length(ζ), 3);
  for ii = 1:3
    c2_err[:,:,ii] = c22[:,:,ii] - c2[:,:,ii] - Diagonal(dα)*dc2[:,:,1,ii] - dc2[:,:,2,ii]*Diagonal(dζ)
  end
  bsuppar_err = bsuppar2 - bsuppar - Diagonal(dα)*dbsuppar[:,:,1] - dbsuppar[:,:,2]*Diagonal(dζ)

  println("FINITE DIFFERENCE CHECK c2, κ, AND bsuppar
  error in c2: $(norm(c2_err)/ϵ)
  error in bsuppar: $(norm(bsuppar_err)/ϵ)
  ")

  #
  x = 0:1:2
  nx = length(x)
  ddcoeff_data_1 = ddcoeffs_struct(nx);
  ddcoeff_data_1.x[:] .= x;
  ddcoeff_data_2 = ddcoeffs_struct(nx);
  ddcoeff_data_2.x[:] .= x;
  coeff_data = coeffs_struct(nx);
  coeff_data.x[:] .= x;

  coeffs(bci, r[2], r[1], coeff_data)
  coeffs_derivs!(bci, r[2], r[1], ddcoeff_data_1)
  coeffs_derivs!(bci, r2[2], r2[1], ddcoeff_data_2)

  println("CHECK THAT p, q, AND r MATCH coeffs()
  error in p: $(norm(coeff_data.p - ddcoeff_data_1.p))
  error in q: $(norm(coeff_data.q - ddcoeff_data_1.q))
  error in r: $(norm(coeff_data.r - ddcoeff_data_1.r))
  ")

  println("FINITE DIFFERENCE CHECK norm(var2 - var1 - dvardr * dr)/eps
    p: $(norm(ddcoeff_data_2.p - ddcoeff_data_1.p - ddcoeff_data_1.dp*dr)/ϵ)
    q: $(norm(ddcoeff_data_2.q - ddcoeff_data_1.q - ddcoeff_data_1.dq*dr)/ϵ)
    r: $(norm(ddcoeff_data_2.r - ddcoeff_data_1.r - ddcoeff_data_1.dr*dr)/ϵ)
    ")


  rel_err = (dp2, dp1, ddp) -> norm(vec(dp2)-vec(dp1)-reshape(ddp,2nx,2)*dr)/ϵ
  println("SECOND FINITE DIFFERENCE CHECK norm(dvar2 - dvar1 - ddvardr * dr)/eps
    p: $(rel_err(ddcoeff_data_2.dp,ddcoeff_data_1.dp,ddcoeff_data_1.ddp))
    q: $(rel_err(ddcoeff_data_2.dq,ddcoeff_data_1.dq,ddcoeff_data_1.ddq))
    r: $(rel_err(ddcoeff_data_2.dr,ddcoeff_data_1.dr,ddcoeff_data_1.ddr))
    ")
end


function adaptive_test(s, N1, N2)
  vmec = VMEC.readVmecWout("../example/wout_ku4.nc");
  vmecSurface = VmecSurface(s, vmec);
  bci = BallooningCoefInterp(vmecSurface, N1, N2)

  c2_coef = get_c2(bci);
  κ_coef = get_κ(bci);
  bsuppar_coef = get_bsuppar(bci);


  return c2_coef, κ_coef, bsuppar_coef
end

function cobra_test(s)
  vmec = VMEC.readVmecWout("../example/wout_ku4.nc");
  # θ =
end
