function dλ_test()
    n = 4;
    dv = randn(n);
    de = randn(n-1);
    A = SymTridiagonal(dv, de);

    M_on = true
    M = M_on ? Diagonal(abs.(randn(n)) .+ 1) : Diagonal(ones(n));

    dA1_on = 1;
    dA2_on = 1;
    dA = [SymTridiagonal(dA1_on .* randn(n), dA1_on .* randn(n-1)),
          SymTridiagonal(dA2_on .* randn(n), dA2_on .* randn(n-1))];

    dM1_on = 1;
    dM2_on = 1;
    dM = [Diagonal(dM1_on .* randn(n)),
          Diagonal(dM2_on .* randn(n))];

    ddA11_on = 1;
    ddA12_on = 1;
    ddA22_on = 1;
    ddA = [SymTridiagonal(ddA11_on .* randn(n), ddA11_on .* randn(n-1)),
           SymTridiagonal(ddA12_on .* randn(n), ddA12_on .* randn(n-1)),
           SymTridiagonal(ddA12_on .* randn(n), ddA12_on .* randn(n-1)),
           SymTridiagonal(ddA22_on .* randn(n), ddA22_on .* randn(n-1))];
    ddA = reshape(ddA, 2, 2);
    ddA[1,2] = ddA[2,1];

    ddM11_on = 1;
    ddM12_on = 1;
    ddM22_on = 1;
    ddM = [Diagonal(ddM11_on .* randn(n)),
           Diagonal(ddM12_on .* randn(n)),
           Diagonal(ddM12_on .* randn(n)),
           Diagonal(ddM22_on .* randn(n))];
    ddM = reshape(ddM, 2, 2);
    ddM[1,2] = ddM[2,1];


    A_fn = (r) -> begin
        A_eval = A + dA'*r + (r'*ddA*r)/2;
        dA_eval = ddA*r .+ dA;
        ddA_eval = deepcopy(ddA);
        (A_eval, dA_eval, ddA_eval)
    end

    M_fn = (r) -> begin
        M_eval = M + dM'*r + (r'*ddM*r)/2;
        dM_eval = ddM*r .+ dM;
        ddM_eval = deepcopy(ddM);
        (M_eval, dM_eval, ddM_eval)
    end

    pt = [0.,0];
    A1, dA1, ddA1 = A_fn(pt);
    println("typeof(A1) = $(typeof(A1))")
    M1, dM1, ddM1 = M_fn(pt);
    λ1, x1 = get_λx(A1, M1)
    println("Eigenvalue at $pt is λ = $λ1")

    eps = 1e-5;
    dpt = eps .* randn(2);
    A2, dA2, ddA2 = A_fn(pt + dpt);
    M2, dM2, ddM2 = M_fn(pt + dpt);
    λ2, x2 = get_λx(A2, M2);
    println("Eigenvalue at $(pt+dpt) is λ = $λ2")

    dλ1, dx1, ddλ1, ddx1 = get_λx_derivs(A1, M1, dA1, dM1, ddA1, ddM1, λ1, x1)
    dλ2, dx2, ddλ2, ddx2 = get_λx_derivs(A2, M2, dA2, dM2, ddA2, ddM2, λ2, x2)

    # println("A1: "); flush(stdout)
    # display(A1);
    # println("dA1: "); flush(stdout)
    # display(dA1);
    # println("ddA1: "); flush(stdout)
    # display(ddA1);
    # println("A2-A1: "); flush(stdout)
    # display(A2-A1);
    # println("dA1'*dpt: "); flush(stdout)
    # display(dA1'*dpt);
    # println("A2 - A1 - dA1'*dpt: "); flush(stdout)
    # display(A2 - A1 - dA1'*dpt);

    # println("dx2-dx1"); flush(stdout)
    # display(dx2-dx1)
    # println("ddx1*dpt"); flush(stdout)
    # display(reshape(reshape(ddx1, 2n, 2)*dpt, n, 2))

    println("x1'*M1*x1 = $((x1'*M1*x1))")
    println("x2'*M2*x2 = $((x2'*M2*x2))")
    println("norm((A1 - λ1 * M1)*x1) = $(norm((A1 - λ1 * M1)*x1))")
    println("norm((A2 - λ2 * M2)*x2) = $(norm((A2 - λ2 * M2)*x2))")
    println("x1'*M1*(x2-x1) = $(x1'*M1*(x2-x1))")
    println("dA finite difference test: $(norm(A2 - A1 - dA1'*dpt)/eps)")
    println("dM finite difference test: $(norm(M2 - M1 - dM1'*dpt)/eps)")
    println("dλ finite difference test: $(norm(λ2 - λ1 - dλ1'*dpt)/eps)")
    println("dx finite difference test: $(norm(x2 - x1 - dx1*dpt)/eps)")
    println("ddA finite difference test: $(norm(norm.(dA2-dA1-ddA1*dpt))/eps)")
    println("ddM finite difference test: $(norm(norm.(dM2-dM1-ddM1*dpt))/eps)")
    println("ddλ finite difference test: $(norm(dλ2 - dλ1 - ddλ1'*dpt)/eps)")
    println("ddx finite difference test: $(norm(dx2[:] - dx1[:] - reshape(ddx1,2n,2)*dpt)/eps)")

end

function eig_opt_test()
    nfp = 3;
    Afun = (r) -> begin
        θ = r[1]; ζ = r[2];
        dv = [cos(θ)*sin(nfp*ζ), -5, -5];
        de = [1., 1.];
        ∇dv = [[-sin(θ)*sin(nfp*ζ), 0, 0], [nfp*cos(θ)*cos(nfp*ζ), 0, 0]];
        ∇ev = [[0., 0.], [0., 0.]];
        ∇∇dv = reshape([[-cos(θ)*sin(nfp*ζ), 0, 0],
                        [-nfp*sin(θ)*cos(nfp*ζ), 0, 0],
                        [-nfp*sin(θ)*cos(nfp*ζ), 0, 0],
                        [-nfp^2*cos(θ)*sin(nfp*ζ), 0, 0]], 2, 2);
        ∇∇ev = reshape([[0., 0.], [0., 0.], [0., 0.], [0., 0.]], 2, 2);

        A = SymTridiagonal(dv, de);
        dA = [SymTridiagonal(∇dv[ii], ∇ev[ii]) for ii = 1:2];
        ddA = [SymTridiagonal(∇∇dv[ii, jj], ∇∇ev[ii, jj]) for ii = 1:2, jj = 1:2];

        (A, dA, ddA)
    end

    Mfun = (r) -> begin
        M = Diagonal(ones(3));
        dM = [Diagonal(zeros(3)) for ii = 1:2]
        ddM = [Diagonal(zeros(3)) for ii = 1:2, jj = 1:2]

        (M, dM, ddM)
    end

    r0 = abs.(randn(2));

    r, λ = eig_opt(Afun, Mfun, r0, nfp)
    println("λ = $λ, r = $r")
end


function real_eig_opt_test()
    s = 3/11;
    # vmec = VMEC.readVmecWout("../example/wout_ku4.nc");
    vmec = VMEC.readVmecWout("../example/wout_wista_15_beta_0.nc");
    vmecSurface = VmecSurface(s, vmec);

    nwells = 10;
    geom_input = true;
    tokamak_input=false;
    mode=1;
    vmec_data = VmecData(vmec)
    (r0, amin, beta0, b0_v) = order_input(vmec_data);
    params = BalloonParams(nwells, geom_input, tokamak_input, mode, r0, amin, beta0, b0_v)

    bci = BallooningCoefInterp(vmecSurface, params)
    nfp = vmecSurface.nfp;


    np0_in = 793;
    xmax = π*(2*params.k_w - 1) ./ (2*vmecSurface.nfp*vmecSurface.iota[1])
    np  = Int(6 * floor(np0_in / 6) + 1);      # number of points always odd and multiple of 3
    xmin = -xmax;
    h   = 2*(xmax/(np - 1));
    AMfun = getmatrix_function(xmin, h, np, bci)

    pt = [0.3, 0.3];
    A1, dA1, ddA1, M1, dM1, ddM1 = AMfun(pt);

    eps = 1e-5;
    dpt = eps .* randn(2);
    A2, dA2, ddA2, M2, dM2, ddM2 = AMfun(pt + dpt);

    println("dA finite difference test: $(norm(A2 - A1 - dA1'*dpt)/(eps*norm(A1)))")
    println("dM finite difference test: $(norm(M2 - M1 - dM1'*dpt)/(eps*norm(M1)))")
    println("ddA finite difference test: $(norm(norm.(dA2-dA1-ddA1*dpt))/(eps*norm(norm.(dA1))))")
    println("ddM finite difference test: $(norm(norm.(dM2-dM1-ddM1*dpt))/(eps*norm(norm.(dM1))))")

    # Check periodicity
    A1, dA1, ddA1, M1, dM1, ddM1 = AMfun([0.0, 0.0]);
    A2, dA2, ddA2, M2, dM2, ddM2 = AMfun([0.0, 2π/get_nfp(bci)]);
    println("PERIODICITY CHECK
  A: $(norm(A1-A2))
  M: $(norm(M1-M2))
  dA: $(norm(norm.(dA1-dA2)))
  dM: $(norm(norm.(dM1-dM2)))
  ddA: $(norm(norm.(ddA1-ddA2)))
  ddM: $(norm(norm.(ddM1-ddM2)))")

  θ0s = (0:4) .* (2π/5)
  ζ0s = (0:4) .* (2π*get_nfp(bci)/5);
  λs = zeros(5,5);
  rmins = zeros(5, 5, 2);
  for ii = 1:5, jj = 1:5
      rmins[ii, jj, :], λs[ii, jj] = eig_opt(AMfun, [θ0s[ii], ζ0s[jj]], get_nfp(bci); tol = 1e-8)
  end

  # α = θ - ι*ζ;
  println("α coordinate:"); flush(stdout);
  display(rmins[:,:,1] - get_iota(bci) .* rmins[:, :, 2])
  println("ζ coordinate:"); flush(stdout);
  display(rmins[:,:,2])
  println("λs:"); flush(stdout);
  display(λs)


end
